/*
 *  Copyright 2002-2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.mime4j;

import java.util.*;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.EOFException;
import java.io.IOException;

/**
 * Parses RFC822/MIME headers.
 */
public class Headers {
    private final HashMap headers;

    private Headers(byte[] bytes) {

        // normalize \r\n to \n
        StringBuffer sb = new StringBuffer(bytes.length);
        int lastNonLFPos = 0;  // use this to omit trailing whitespace
        for (int i = 0; i < bytes.length; i++) {
            switch (bytes[i]) {
                case '\r':
                    break;
                case '\n':
                    if (!(i < bytes.length - 1 && (bytes[i + 1] == ' ' || bytes[i + 1] == '\t')))
                        sb.append((char)bytes[i]);  // only append if folded whitespace not detected
                    break;
                default:
                    sb.append((char)bytes[i]);
                    lastNonLFPos = sb.length();
                    break;
            }
        }
        String headerString = lastNonLFPos == 0 ? "" : sb.substring(0, lastNonLFPos);

        StringTokenizer headerLines;
        if (headerString.length() == 0)
            headerLines = null;
        else
            headerLines = new StringTokenizer(headerString, "\n", false);

        this.headers = new HashMap();

        if (headerLines != null) {
            while (headerLines.hasMoreTokens()) {
                String headerLine = headerLines.nextToken();

                int splitAt = headerLine.indexOf(":");
                if (splitAt == -1) {
                    Debug.fail("Malformed header line: " + headerLine);
                    continue;
                }

                String headerName = headerLine.substring(0, splitAt).toLowerCase();
                int valueIndex = splitAt + 1;
                while (valueIndex < headerLine.length() && (headerLine.charAt(valueIndex) == ' ' || headerLine.charAt(valueIndex) == '\t'))
                    valueIndex++;
                String headerValue = headerLine.substring(valueIndex);

                ArrayList list;
                if (null == (list = (ArrayList)this.headers.get(headerName))) {
                    list = new ArrayList(1);
                    headers.put(headerName, list);
                }
                list.add(headerValue);
            }
        }
    }

    public boolean containsHeader(String headerName) {
        return headers.containsKey(headerName.toLowerCase());
    }

    public String getHeaderValue(String headerName) {
        ArrayList values = (ArrayList)headers.get(headerName.toLowerCase());
        if (values != null)
            return (String)values.get(0);
        else
            return null;
    }

    public String[] getHeaderValues(String headerName) {
        ArrayList values = (ArrayList)headers.get(headerName.toLowerCase());

        if (values == null)
            return null;

        String[] result = new String[values.size()];
        result = (String[])values.toArray(result);
        return result;
    }

    public Map getHeaderParams(String headerName, boolean paramNamesToLowerCase) {
        String headerValue = getHeaderValue(headerName);
        if (headerValue == null)
            return null;
        Map result = new Hashtable();

        // split main value and parameters
        String main;
        String rest;
        if (headerValue.indexOf(";") == -1) {
            main = headerValue;
            rest = null;
        } else {
            main = headerValue.substring(0, headerValue.indexOf(";"));
            rest = headerValue.substring(main.length() + 1);
        }

        result.put("", main);
        if (rest != null) {
            char[] chars = rest.toCharArray();
            StringBuffer paramName = new StringBuffer();
            StringBuffer paramValue = new StringBuffer();

            final byte READY_FOR_NAME = 0;
            final byte IN_NAME = 1;
            final byte READY_FOR_VALUE = 2;
            final byte IN_VALUE = 3;
            final byte IN_QUOTED_VALUE = 4;
            final byte VALUE_DONE = 5;
            final byte ERROR = 99;

            byte state = READY_FOR_NAME;
            for (int i = 0; i < chars.length; i++) {
                char c = chars[i];

                switch (state) {
                    case ERROR:
                        if (c == ';')
                            state = READY_FOR_NAME;
                        break;

                    case READY_FOR_NAME:
                        if (c == '=') {
                            Debug.fail("Expected header param name, got '='");
                            state = ERROR;
                            break;
                        }

                        paramName = new StringBuffer();
                        paramValue = new StringBuffer();

                        state = IN_NAME;
                        // fall-through

                    case IN_NAME:
                        if (c == '=') {
                            if (paramName.length() == 0)
                                state = ERROR;
                            else
                                state = READY_FOR_VALUE;
                            break;
                        }

                        // not '='... just add to name
                        paramName.append(c);
                        break;

                    case READY_FOR_VALUE:
                        boolean fallThrough = false;
                        switch (c) {
                            case ' ':
                            case '\t':
                                break;  // ignore spaces, especially before '"'

                            case '"':
                                state = IN_QUOTED_VALUE;
                                break;

                            default:
                                state = IN_VALUE;
                                fallThrough = true;
                                break;
                        }
                        if (!fallThrough)
                            break;

                        // fall-through

                    case IN_VALUE:
                        switch (c) {
                            case ';':
                                result.put(
                                        paramNamesToLowerCase ? paramName.toString().trim().toLowerCase() : paramName.toString().trim(),
                                        paramValue.toString().trim());
                                state = VALUE_DONE;
                                break;

                            default:
                                paramValue.append(c);
                                break;
                        }
                        break;

                    case IN_QUOTED_VALUE:
                        switch (c) {
                            case '"':
                                // don't trim quoted strings; the spaces could be intentional.
                                result.put(
                                        paramNamesToLowerCase ? paramName.toString().trim().toLowerCase() : paramName.toString().trim(),
                                        paramValue.toString());
                                state = VALUE_DONE;
                                break;

                            default:
                                paramValue.append(c);
                                break;
                        }
                        break;

                    case VALUE_DONE:
                        switch (c) {
                            case ';':
                                state = READY_FOR_NAME;
                                break;

                            case ' ':
                            case '\t':
                                break;

                            default:
                                state = ERROR;
                                break;
                        }
                        break;
                }
            }

            // done looping.  check if anything is left over.
            if (state == IN_VALUE) {
                result.put(
                        paramNamesToLowerCase ? paramName.toString().trim().toLowerCase() : paramName.toString().trim(),
                        paramValue.toString().trim());
            }
        }

        return result;
    }

    /**
     * Parses headers from the given stream.
     *
     * Assumes that the stream is positioned at the beginning of a block
     * of headers.  After this method's successful return, the stream will
     * be positioned immediately following the first [\r]\n[\r]\n encountered.
     * @throws IOException Underlying stream threw exception.
     * @throws EOFException EOF occurred before headers ended.
     */
    public static Headers read(InputStream s) throws IOException {
        ByteArrayOutputStream ms = new ByteArrayOutputStream(1024);

        int state = 1;
        for (; ;) {
            int b = s.read();
            switch (b) {
                case -1:
                    throw new EOFException();
                case 13:
                    ms.write(b);
                    break;
                case 10:
                    ms.write(b);
                    if (state == 0)
                        state = 1;
                    else if (state == 1)
                        return new Headers(ms.toByteArray());
                    break;
                default:
                    ms.write(b);
                    state = 0;
                    break;
            }
        }
    }

}
