/*
 *  Copyright 2002-2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.mime4j;

import java.io.InputStream;
import java.io.IOException;

/**
 * Stream that constrains itself to a single MIME body part.
 *
 * After the stream ends (i.e. ReadByte() returns -1 or Read()
 * returns 0), the BoundaryState property can be read to see
 * what caused the stream to end--normal boundary, final boundary,
 * or (unexpected) EOF.
 */
public class MIMEBoundaryInputStream extends InputStream {

    /**
     * Start of stream; possible immediate boundary
     */
    public static final int STATE_BEGIN = 0;

    /**
     * In stream; expect \n before boundary
     */
    public static final int STATE_NORMAL = 1;

    /**
     * Hit regular boundary; stream is done
     */
    public static final int STATE_BOUNDARY = 2;

    /**
     * Hit final boundary; stream is done
     */
    public static final int STATE_BOUNDARY_FINAL = 3;

    /**
     * hit unexpected EOF; stream is done
     */
    public static final int STATE_EOF = 4;

    private final InputStream s;
    private final byte[] boundary;
    private final int desiredLookahead;
    private final ByteQueue lookahead = new ByteQueue();
    private int state = STATE_BEGIN;

    /**
     * Creates a new MIMEBoundaryInputStream.
     * @param s The underlying stream.
     * @param boundary Boundary bytes (not including leading hyphens).
     */
    public MIMEBoundaryInputStream(InputStream s, byte[] boundary) {
        this.s = s;

        this.boundary = new byte[2 + boundary.length];
        this.boundary[0] = this.boundary[1] = (byte)'-';
        System.arraycopy(boundary, 0, this.boundary, 2, boundary.length);

        desiredLookahead = this.boundary.length;
    }

    /**
     * Creates a new MIMEBoundaryInputStream.
     * @param s The underlying stream.
     * @param boundary Boundary string (not including leading hyphens).
     */
    public MIMEBoundaryInputStream(InputStream s, String boundary) {
        this.s = s;

        boundary = "--" + boundary;
        this.boundary = new byte[boundary.length()];
        desiredLookahead = this.boundary.length;
        for (int i = 0; i < desiredLookahead; i++) {
            this.boundary[i] = (byte)boundary.charAt(i);
        }
    }

    /**
     * Gets the current state.  Returns one of the STATE_ constants.
     */
    public int getState() {
        return state;
    }

    public int read() throws IOException {
        lookahead();

        if (lookahead.count() == 0)
            return -1;
        else
            return (byte)lookahead.dequeue();

    }

    private int doRead() throws IOException {
        int i = s.read();
        if (i == -1)
            state = STATE_EOF;
        return i;
    }

    /**
     * Reads some characters out of the underlying stream (if available)
     * and into the lookahead buffer.
     *
     * If [\r]\n is read, then the method also tests to see if we are at
     * the MIME boundary.  If not, all the characters that were consumed
     * during the test are also added to the lookahead.  If so, the boundary
     * characters are discarded and no more characters will be read from
     * the stream.
     * @throws IOException Underlying stream threw an exception.
     */
    private void lookahead() throws IOException {
        boolean tryMatch = false;  // need to test for MIME boundary?

        // need to save these in case MIME test returns false
        int c1 = -1;  // 1st character consumed during tryMatch test
        int c2 = -1;  // 2nd character consumed during tryMatch test

        switch (state) {
            case STATE_BOUNDARY:
            case STATE_BOUNDARY_FINAL:
            case STATE_EOF:
                return;  // nothing more to see

            case STATE_BEGIN:  // we always need to try to match
                tryMatch = true;
                state = STATE_NORMAL;
                break;

            case STATE_NORMAL:
                int i = doRead();
                c1 = i;
                if (i == -1) {
                    state = STATE_EOF;
                    return;
                } else if (i == '\r') {
                    i = doRead();
                    c2 = i;
                    if (i == '\n')
                        tryMatch = true; // \r\n matched, try to match
                    else
                        Debug.fail("Encountered \\r without \\n");
                } else if (i == '\n') {
                    tryMatch = true; // \n without preceding \r; bad MIME!
                }

                break;
            default:
                Debug.fail("Unhandled State " + state);
                return;
        }

        if (tryMatch) {
            int i;  // iteration var
            int c = -1;  // current character
            for (i = 0; i < boundary.length; i++) {
                c = doRead();
                if (c != boundary[i])  // includes case where c == -1
                    break;
            }
            if (i == boundary.length) {
                // full match!  see if we're at the end
                if (readToBoundaryEOL()) {
                    state = STATE_BOUNDARY_FINAL;
                } else {
                    state = STATE_BOUNDARY;
                }
            } else {
                // wasn't a match; add consumed characters to lookahead

                // first add original 0, 1, or 2 characters that
                // told us we needed to try a match
                if (c1 != -1) {
                    lookahead.enqueue((byte)c1);
                    if (c2 != -1)
                        lookahead.enqueue((byte)c2);
                }

                // then add the characters that matched
                for (int j = 0; j < i; j++)
                    lookahead.enqueue(boundary[j]);

                // finally, add the character that didn't match
                if (c != -1)
                    lookahead.enqueue((byte)c);
            }
        } else {
            lookahead.enqueue((byte)c1);
            if (c2 != -1)
                lookahead.enqueue((byte)c2);
        }
    }

    /**
     * Read to end of line or end of file.
     * @return true if first two chars are hyphens.
     * @throws IOException Underlying stream threw an exception.
     */
    private boolean readToBoundaryEOL() throws IOException {
        int charsRead = 0;
        byte hyphensRead = 0;

        int b;
        while (-1 != (b = s.read())) {
            if (b == '\n')
                break;

            if (b == '-' && charsRead == 0) {
                hyphensRead = 1;
            } else if (b == '-' && charsRead == 1 && hyphensRead == 1) {
                hyphensRead = 2;
            }

            charsRead++;
        }

        return hyphensRead == 2;
    }


}
