/*
 *  Copyright 2002-2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.mime4j;

import java.io.InputStream;
import java.io.IOException;
import java.util.Map;

/**
 * Parses a MIME e-mail message.
 *
 * To use, create a subclass of this class, and override the
 * relevant OnXxxx methods.  Then instantiate the subclass
 * using the raw message stream, and call Parse().
 */
public class MIMEParser {

    private final InputStream s;
    private final MIMEEventCallbacks callbacks;
    private final Options options;

    public MIMEParser(InputStream s, MIMEEventCallbacks callbacks) {
        this(s, callbacks, new Options());
    }

    public MIMEParser(InputStream s, MIMEEventCallbacks callbacks, Options options) {
        this.s = s;
        this.callbacks = callbacks;
        this.options = options;
    }

    public void parse() throws IOException {
        matchHeader(s, true);
    }

    protected void matchHeader(InputStream s, boolean needsMimeVersion) throws IOException {
        Headers headers = Headers.read(s);
        boolean isMime = !needsMimeVersion || headers.containsHeader("MIME-Version");

        callbacks.onHeaders(headers);
        if (isMime)
            matchBody(s, headers);
        else
            callbacks.onNonMimeBody(headers, new CloseShieldInputStream(s));
    }

    protected void matchBody(InputStream s, Headers headers) throws IOException {
        boolean isMultiPart = false;
        boolean isMessageRfc822 = false;
        String boundary = null;
        if (headers.containsHeader("Content-Type")) {
            Map contentTypeParams = headers.getHeaderParams("Content-Type", true);
            if (contentTypeParams.containsKey("")) {
                String contentType = (String)contentTypeParams.get("");
                if (contentType != null) {
                    contentType = contentType.trim().toLowerCase();
                    if (contentType.startsWith("multipart/")) {
                        boundary = (String)contentTypeParams.get("boundary");
                        if (boundary != null)
                            isMultiPart = true;
                    } else if (contentType.startsWith("message/rfc822") && options.getProcessNestedMessages()) {
                        isMessageRfc822 = true;
                    }
                }
            }
        }

        if (isMultiPart) {
            matchMultipart(s, headers, boundary);
            skipToEnd(s);
        } else {
            InputStream filteredStream = s;
            if (headers.containsHeader("Content-Transfer-Encoding")) {
                String encoding = headers.getHeaderValue("Content-Transfer-Encoding").toLowerCase();
                if (encoding.equals("quoted-printable") && options.getDecodeQuotedPrintable())
                    filteredStream = new QuotedPrintableInputStream(s);
                else if (encoding.equals("base64") && options.getDecodeBase64())
                    filteredStream = new Base64InputStream(s);
            }

            if (isMessageRfc822) {
                callbacks.onMessageStart(headers);
                matchHeader(filteredStream, true);
                skipToEnd(s);
                callbacks.onMessageEnd();
            } else {
                callbacks.onDiscreteBodyPart(headers, new CloseShieldInputStream(filteredStream));
                skipToEnd(s);
            }
        }

    }

    protected void matchMultipart(InputStream s, Headers headers, String boundary) throws IOException {
        callbacks.onMultipartStart(headers);

        MIMEBoundaryInputStream bs = new MIMEBoundaryInputStream(s, boundary);
        callbacks.onPreamble(new CloseShieldInputStream(bs));
        skipToEnd(bs);

        while (bs.getState() == MIMEBoundaryInputStream.STATE_BOUNDARY) {
            bs = new MIMEBoundaryInputStream(s, boundary);
            matchHeader(bs, false);
        }

        callbacks.onEpilogue(new CloseShieldInputStream(s));
        skipToEnd(s);

        callbacks.onMultipartEnd();
    }

    private void skipToEnd(InputStream s) throws IOException {
        while (s.read() != -1);
    }



    public static class Options {

        private boolean decodeBase64 = true;
        private boolean decodeQuotedPrintable = true;
        private boolean processNestedMessages = true;

        public Options() {
        }

        public boolean getDecodeBase64() {
            return decodeBase64;
        }

        /**
         * Set to false to prevent base64 streams from being decoded.
         * Instead, the raw encoded streams will be given.
         * Defaults to true.
         */
        public void setDecodeBase64(boolean decodeBase64) {
            this.decodeBase64 = decodeBase64;
        }

        public boolean getDecodeQuotedPrintable() {
            return decodeQuotedPrintable;
        }

        /**
         * Set to false to prevent quoted-printable streams from being decoded.
         * Instead, the raw encoded streams will be given.
         * Defaults to true.
         */
        public void setDecodeQuotedPrintable(boolean decodeQuotedPrintable) {
            this.decodeQuotedPrintable = decodeQuotedPrintable;
        }

        public boolean getProcessNestedMessages() {
            return processNestedMessages;
        }

        /**
         * Set to false to treat nested RFC822 messages the same as any other
         * body part.  Defaults to true.
         */
        public void setProcessNestedMessages(boolean processNestedMessages) {
            this.processNestedMessages = processNestedMessages;
        }
    }
}
