package org.mime4j.headers;

public class FieldParser {

    private byte[] buffer;
    private int mark = -1;
    private int pos;

    private void mark() {
        mark = pos;
    }

    private byte[] getMarkedBytes() {
        int len = pos - mark;
        if (len <= 0) {
            return new byte[0];
        } else {
            byte[] bytes = new byte[len];
            System.arraycopy(buffer, mark, bytes, 0, pos - mark);
            return bytes;
        }
    }

    private boolean matchQuotedString() {
        int state = 0;

        while (!atEol()) {
            byte c = la();
            switch (state) {
                case 0:
                    // haven't started yet
                    if (c != '"')
                        return false;
                    else {
                        state = 1;
                        consumeChar();
                        break;
                    }
                case 1:
                    if (c == '"') {
                        consumeChar();
                        return false;
                    }
                    else if (c == '\\') {
                        // quoted pair
                        consumeChar();
                        if (atEol())
                            return unexpectedEol();
                        consumeChar();
                        break;
                    }
                    else {
                        consumeChar();
                        break;
                    }
            }
        }
        return unexpectedEol();
    }

    private boolean unexpectedEol() {
        // possibly make this throw in the future
        return true;
    }

    private boolean matchDotAtom() {
        // first char cannot be dot
        if (!atEol() && isAtomChar(la())) {
            consumeChar();

            while (!atEol() && (isAtomChar(la()) || la() == '.'))
                consumeChar();
        }
        return atEol();
    }

    private boolean matchAtom() {
        while (!atEol() && isAtomChar(la()))
            consumeChar();
        return atEol();
    }

    private boolean matchCfws() {
        int state = 0;
        int commentNestCount = 0;

        while (!atEol()) {
            byte c = la();
            switch (state) {
                case 0:
                    // look for space or (
                    if (isWhitespace(c))
                        consumeChar();
                    else if (c == '(') {
                        consumeChar();
                        state = 1;
                    }
                    else {
                        return false;
                    }
                    break;

                case 1:
                    // in a comment
                    if (c == '(') {
                        commentNestCount++;
                    }
                    else if (c == ')') {
                        if (commentNestCount == 0) {
                            state = 0;
                            consumeChar();
                        }
                        else
                            commentNestCount++;
                    }
                    else if (c == '\\') {
                        // potential quoted-pair
                        boolean isQuotedPair = hasMore(2);
                        if (!isQuotedPair) {
                            return false;
                        }
                        else {
                            // skip quoted pair
                            consumeChar();
                            consumeChar();
                        }
                    }
                    else {
                        consumeChar();
                    }
            }
        }

        if (state == 1)
            return unexpectedEol();
        else
            return true;
    }

    /**
     * Consumes whitespace characters.
     *
     * @return true if EOF reached
     */
    private boolean matchWhitespace() {
        while (!atEol() && isWhitespace(la()))
            consumeChar();
        return atEol();
    }

    private void consumeChar() {
        pos++;
    }

    private boolean isAtomChar(byte b) {

        if (b >= 'a' && b <= 'z')
            return true;
        if (b >= 'A' && b <= 'Z')
            return true;
        if (b >= '0' && b <= '9')
            return true;

        switch (b) {
            case '!':
            case '#':
            case '$':
            case '%':
            case '&':
            case '\'':
            case '*':
            case '+':
            case '-':
            case '/':
            case '=':
            case '?':
            case '^':
            case '_':
            case '`':
            case '{':
            case '|':
            case '}':
            case '~':
                return true;
            default:
                return false;
        }
    }

    private boolean isWhitespace(byte b) {
        switch (b) {
            case (byte)' ':
            case (byte)'\t':
                return true;
            default:
                return false;
        }
    }

    private boolean atEol() {
        return pos >= buffer.length;
    }

    private boolean hasMore(int count) {
        return pos + count - 1 < buffer.length;
    }

    /**
     * look at the current char
     */
    private byte la() {
        return buffer[pos];
    }
}
